#!/usr/bin/python


import sys
import os.path
import socket, copy
import pprint, textwrap
from xml.dom import minidom
import zipfile
import tempfile
import re
import shutil
import cStringIO

sys.path.append('/usr/share/rhn')
from optparse import Option, OptionParser
try:
    from common import rhnLib, rhnLog, initLOG, CFG, initCFG
    from server import rhnSQL
except ImportError:
    from spacewalk.common.rhnConfig import CFG, initCFG
    from spacewalk.common import rhnLib, rhnLog
    from spacewalk.common.rhnLog import initLOG
    from spacewalk.server import rhnSQL

import recognize_guests



verbose = 0
debug = 0
sat_version = 0

def error_exit(msg):
    sys.stderr.write(msg);
    sys.stderr.write("\n");
    sys.exit(-1);

def main():
    global verbose, debug, sat_version

    initCFG('web')
    sat_version = get_sat_version()

    initCFG('server.satellite')
    initLOG(CFG.LOG_FILE, CFG.DEBUG)

    verbose = 0
    debug = 0

    options_table = [
        Option("-c", "--rhn-cert", action="store",
               help="The rhn certificate to activate"),
        Option("-m", "--manifest", action="store",
               help="Path to manifest file"),
        Option("-v", "--verbose", action="count",
               help="Increase verbosity"),
        Option("--debug", action="store_true",
               help="logs the debug information to a log file"),
    ]

    parser = OptionParser(option_list=options_table)
    (options, args) = parser.parse_args()
    if args:
        for arg in args:
            error_exit("Not a valid option ('%s'), try --help\n" % arg)


    if options.verbose:
        verbose = 1
    if options.debug:
        debug = 1
    if not options.rhn_cert and not options.manifest:
        error_exit("Missing --rhn-cert FILENAME or --manifest FILENAME option")
    if options.rhn_cert and options.manifest:
        error_exit("Please enter only rhn-cert or manifest, not both")

    rhnSQL.initDB()

    #see if any systems can be converted to guests.  Recommend doing this
    (to_convert, cant_convert) = recognize_guests.get_guest_lists()
    if to_convert or cant_convert:
        print("WARNING: %s physical systems have been detected which may be virtual machines.\n\
                 It is recommended you use recognize_guests.py to convert these systems.\n\
                 The results of this script may be inaccurate until this is done.\n" % (len(to_convert) + len(cant_convert)))

    results = True
    #Try to parse the cert
    if options.manifest:
        cert = parse_manifest(options.manifest)
    else:
        cert = parse_cert(options.rhn_cert)

    for family, (quantity, flex)  in cert.items():
        quantity = quantity - flex
        if verbose:
            print("Checking %s: %s,%s" %(family, quantity, flex))
        results = check_channel_family(family, quantity, flex) and  results

    if results:
        print("Upgrade entitlement test PASSED")
    else:
        print("Upgrade entitlement test FAILED")


def check_channel_family(label, quantity, flex):
    return global_test(label, quantity, flex) and org_test(label, quantity, flex)

def org_test(label, quantity, flex):
    #5.3 and older didn't know about flex guest, so treat it as 0
    if sat_version > 5.3:
        flex_sql = "CFO.fve_current_members, coalesce(CFO.fve_max_members, 0) as fve_max_members,"
    else:
        flex_sql = "0 as fve_current_members, 0 as fve_max_members,"

    #selects the org id and system counts.  Physical systems, virtual systems
    #  that are not getting a free entitlement for the current channel family
    org_counts_sql = """
        select CFO.org_id, CFO.current_members, coalesce(CFO.max_members, 0) as max_members, CFO.label,
           """ + flex_sql + """
           (select count(CFS.server_id)
            from rhnChannelFamilyServers CFS
            where CFS.customer_id = CFO.org_id
              and CFS.channel_family_id = CFO.id
              and CFS.server_id not in
                 (select virtual_system_id from rhnVirtualInstance)) as physical_count,
            (select count(CFS.server_id)
            from rhnChannelFamilyServers CFS
            where CFS.customer_id = CFO.org_id
              and CFS.channel_family_id = CFO.id
              and CFS.server_id in
                 (select virtual_system_id from rhnVirtualInstance)
              and CFS.server_id not in
                 (select SV.server_id from rhnChannelFamilyServerVirtual SV
                                       where SV.CHANNEL_FAMILY_ID = CFO.id)) as virtual_count
        from  rhnChannelFamilyOverview CFO
        where  CFO.label = :label
    """
    orgs = rhnSQL.execute(org_counts_sql, label=label).fetchall_dict() or []

    org_1 = {'physical_regular_sys':0, 'flex_regular_sys':0, 'flex_current':0,
                  'flex_max':0, 'regular_current':0, 'regular_max':0}
    totals = copy.deepcopy(org_1)


    for org in orgs:
        if org['org_id'] == 1:
            __add__(org, org_1) #We want *just* org 1
        else:
            __add__(org,totals) #but everything else should be added

    if org_1['physical_regular_sys'] > quantity - totals['regular_max']:
        if verbose > 1:
             print("%s >= %s - %s" % (org_1['physical_regular_sys'], quantity, totals['regular_max']))
        format_print("""ERROR: You have too many physical systems for the %s channel family.  There are %s physical systems
                        in organization 1, and %s physical entitlements allocated to all other organizations, while there are only
                        %s entitlements available. """ % (label, org_1['physical_regular_sys'], totals['regular_max'], quantity))

        #if we have more allocated to other orgs than we have in the cert
        if totals['regular_max'] > quantity:
            format_print("You will need to move %s physical entitlements from secondary organizations into\
                   the first org in order to be able to activate this certificate." % (totals['regular_max'] - quantity))
        else: #else, we have enough in org 1, but there are too many systems in org 1
            format_print("You will need to unentitle %s systems within the first org, or move\
                   %s entitlements from secondary orgs to the first org." % (org_1['physical_regular_sys'] -\
                   (quantity - totals['regular_max'])))
        suggest_new_cert(org_1, totals, quantity, flex)
        print("") #newline
        return False

    if org_1['physical_regular_sys'] + org_1['flex_regular_sys'] + org_1['flex_current'] >\
              (quantity - totals['regular_max']) + (flex - totals['flex_max']):
        if verbose:
            print("%s + %s + %s > (%s - %s) + (%s - %s)")
        format_print("""ERROR: You have too many physical and guest systems for the %s channel family.
                             You will need to unentitle some systems or purchase more entitlements.""" % (label) )
        suggest_new_cert(org_1, totals, quantity, flex)
        return False

    return True


#Should we suggest the customer call GSS for a new certificate?
def suggest_new_cert(org, totals, quantity, flex):
    if totals['flex_regular_sys'] > 0 and flex > 0:
        format_print("""Alternatively, it is recommended you contact Global Support Services
                          to recieve a temporary certificate to aid in the upgrade process.""")


#private helper to add a map into another map
def __add__(org, totals):
    totals['physical_regular_sys'] += org['physical_count'] #
    totals['flex_regular_sys'] += org['virtual_count'] - org['fve_current_members']     # This is just guests, consuming regular entitlements
    totals['flex_current'] += org['fve_current_members']    #
    totals['flex_max'] += org['fve_max_members']            #
    totals['regular_current'] += org['current_members']     #
    totals['regular_max'] += org['max_members']             #

#Do a simple test to make sure that the customer has enough entitlements at all
#  If this fails, then the user may be using too many physical machines
def global_test(label, quantity, flex):
    guest_count_sql = """
      select SP.server_id, CF.id as family_id
      from  rhnChannelFamilyServerVirtual SP  inner join
            rhnChannelFamily CF on CF.id = SP.channel_family_id
      where  CF.label = :label
        and  SP.server_id in (select virtual_system_id  from rhnVirtualInstance)
    """
    physical_count_sql = """
      select SP.server_id, CF.id as family_id
      from  rhnChannelFamilyServerPhysical SP  inner join
            rhnChannelFamily CF on CF.id = SP.channel_family_id
      where  CF.label = :label
        and  SP.server_id not in (select virtual_system_id  from rhnVirtualInstance)
    """

    flex_count_sql = """
       select distinct SP.server_id, CF.id as family_id
       from  rhnChannelFamilyServerFVE SP inner join
             rhnChannelFamily CF on CF.id = SP.channel_family_id inner join
             rhnChannelFamilyMembers CFM on CFM.channel_family_id = CF.id inner join
             rhnServerChannel SC on SC.server_id = SP.server_id and CFM.channel_id = SC.channel_id
       where CF.label = :label
    """
    guests = rhnSQL.execute(guest_count_sql, label=label).fetchall_dict()
    physical = rhnSQL.execute(physical_count_sql, label=label).fetchall_dict()

    if sat_version > 5.3:
        flex_sys = rhnSQL.execute(flex_count_sql, label=label).fetchall_dict() or []
    else:
        flex_sys = []

    if not guests:
        guests = []
    if not physical:
        physical = []

    family_id = -1
    if physical:
        family_id = physical[0]['family_id']
    elif guests:
        family_id = guests[0]['family_id']

    #if we have too many physcal machines out right, then we need to just fail
    if len(physical) > quantity:
        need = len(physical) - quantity
        format_print("""%s:  Too many physical machines satellite-wide, %s found, %s maximum. You will need to delete
                        or unentitle %s systems, or purchase %s new subscriptions. %s""" %
             (label, len(physical), quantity, need, need, get_family_link(family_id)))
        print("")
        return False

    #We just have too many systems combined (not necessarily too many physical)
    if len(guests) + len(physical) + len(flex_sys)> quantity  + flex:
        need = len(guests) + len(physical) + len(flex_sys) - (quantity  + flex)
        print("""%s:  Too many total systems satellite-wide, %s physical, %s guests, %s physical limit, %s flex limit.
                 You will need to either delete or unentitle some combination of %s physical and flexible systems. %s""" %
                 (label, len(physical), len(guests), quantity, flex, need, get_family_link(family_id)))
        return False

    return True


def get_family_link(id):
    if verbose:
        return "https://%s/network/systems/system_list/in_channel_family.pxt?cfam_id=%s" % (socket.gethostname(), id)
    return ""

def parse_manifest(filename):
    consumer_export_path = "consumer_export.zip"
    zip = zipfile.ZipFile(os.path.expanduser(filename), "r")
    consumer_export_data = zip.read(consumer_export_path)
    consumer_export_data = cStringIO.StringIO(consumer_export_data)
    consumer_export_dataZip = zipfile.ZipFile(consumer_export_data)
    for file in consumer_export_dataZip.namelist():
        if re.match("export\/extensions\/.*", file):
            cert = consumer_export_dataZip.read(file)
            break
    cert = parse_cert_from_string(cert)

    return cert

def parse_cert(filename):
    try:
        file = open(os.path.expanduser(filename), "r")
        return parse_cert_from_string(file.read())
    except IOError:
        error_exit("Cannot read %s" % filename)

def parse_cert_from_string(input_string):
    error = 0
    cert = {}

    xmldoc = minidom.parseString(input_string)

    for child in xmldoc.childNodes[0].childNodes:
        quant = 0
        flex = 0
        if hasValue(child, 'name') and getValue(child, 'name') == 'satellite-version':
            cert_version = child.childNodes[0].data
            rhn_upgrade_version = '.'.join(get_rhn_upgrade_version().split('.')[0:2])
            if not rhn_upgrade_version == cert_version:
                format_print("""WARNING: versions of the supplied certificate [%s] and the Satellite you are about to
                                upgrade to [%s] do not match.""" % (cert_version, rhn_upgrade_version))
            continue
        if not hasValue(child, 'quantity'):
            continue

        label = str(getValue(child, 'family'))
        try:
           quant = int(getValue(child, 'quantity'))
           if hasValue(child, 'flex'):
               flex = int(getValue(child, 'flex'))
           if not cert.has_key(label):
               cert[label] = (quant, flex)
           else:
               error = 1
               print("%s channel label is duplicated, please check cert!" % label)
        except:
            error = 1
            print("Error adding quantity and flex: %s, %s for %s " %\
                 (getValue(child, 'quantity'),  getValue(child, 'flex'), getValue(child, 'name')))

    if error:
        error_exit("Error Parsing Cert!")
    return cert

#Cert parsing helper functions
def getValue(child, name):
    return child.attributes[unicode(name)].value

def hasValue(child,name):
    if child.attributes == None:
       return 0;
    return child.attributes.keys().count(unicode(name))

def format_print(value):
    print('\n'.join(textwrap.wrap(' '.join(value.split()), 80)))


def get_sat_version():
    #CFG could be a float or a string
    split = str(CFG.VERSION).split('.')
    return float("%s.%s" % (split[0], split[1]))

def get_rhn_upgrade_version():
    rpmq = os.popen('rpm -q rhn-upgrade >&/dev/null && rpm -q --qf="%{VERSION}" rhn-upgrade 2>/dev/null','r')
    while 1:
        line = rpmq.readline()
        if not line:
            break
        return line

if __name__ == '__main__':
    main()
