#!/usr/bin/python

import sys
import os
import time
import shutil
import re, pprint
sys.path.append('/usr/share/rhn')

from optparse import Option, OptionParser

try:
    from common import rhnLog, initLOG, CFG, initCFG, rhnFault
    from server import rhnSQL
    from satellite_tools.progress_bar import ProgressBar
except ImportError:
    from spacewalk.common.rhnConfig import CFG, initCFG
    from spacewalk.common import rhnLog
    from spacewalk.common.rhnLog import initLOG
    from spacewalk.server import rhnSQL
    from spacewalk.satellite_tools.progress_bar import ProgressBar

#Copied from server/rhnVirtualization.py from 5.4
class VirtualizationType:
    PARA        = 'para_virtualized'
    FULLY       = 'fully_virtualized'
    QEMU        = 'qemu'
    HYPERV      = 'hyperv'
    VMWARE      = 'vmware'
    VIRTAGE     = 'virtage'

#copied from backend/server/handlers/xmlrpc/registration.py
# only change is to accept QEMU, KVM, and "RHEV Hypervisor" without a UUID)
def parse_smbios(smbios):
    vendor = smbios.get('smbios.bios.vendor')
    serial = smbios.get('smbios.system.serial', '')
    manufacturer = smbios.get('smbios.system.manufacturer')
    product = smbios.get('smbios.system.product')

    # XXX need to worry about uuid being none for other virt types and
    # available subs check
    uuid = None
    if smbios.has_key('smbios.system.uuid'):
        uuid = smbios['smbios.system.uuid']
        uuid = uuid.replace('-', '')

    if vendor == "QEMU":
        return (VirtualizationType.QEMU, uuid)
    elif manufacturer == 'Red Hat' and product == 'KVM':
        return (VirtualizationType.QEMU, uuid)
    elif (manufacturer == 'Red Hat' and product == 'RHEV Hypervisor'):
        return (VirtualizationType.QEMU, uuid)
    elif manufacturer == 'Microsoft Corporation' and \
            product == 'Virtual Machine':
        if uuid is None:
            uuid = "flex-guest"
        return (VirtualizationType.HYPERV, uuid)
    elif serial.startswith('VMware-'):
        if uuid is None:
            uuid = "flex-guest"
        return (VirtualizationType.VMWARE, uuid)
    elif manufacturer == 'HITACHI' and product.endswith(' HVM LPAR'):
        if uuid is None:
            uuid = "flex-guest"
        return (VirtualizationType.VIRTAGE, uuid)
    else:
        return (None, None)


def main():
    initCFG('server.satellite')
    initLOG(CFG.LOG_FILE, CFG.DEBUG)

    OPTIONS = None
    debug = 0

    options_table = [
        Option("--debug", action="store_true",
            help="logs the debug information to a log file"),
        Option("-p", "--pretend", action="store_true",
            help="Do not commit changes to the database"),
    ]

    parser = OptionParser(option_list=options_table)
    (options, args) = parser.parse_args()
    if args:
        for arg in args:
            sys.stderr.write("Not a valid option ('%s'), try --help\n" % arg)
            sys.exit(-1)

    if options.debug:
        debug = 1

    rhnSQL.initDB()

    (to_convert, kvm_convert) =  get_guest_lists(debug = debug)

    if not to_convert and not kvm_convert:
        print("No systems detected to convert to guests")
        sys.exit(0)

    print "Converting %s systems" % (len(to_convert)+len(kvm_convert))
    for system in to_convert:
        convert_system(system['server_id'], system['virt_type'])
    for system in kvm_convert:
        convert_system(system['server_id'], system['virt_type'])

    if kvm_convert:
        print("WARNING: %s un-recognized KVM and/or Xen guests were detected that had been registered with older clients.  These systems are now recognized as guests, but will not have a host/guest relationship if their host is registered in the future.  If you wish to add this relationship, these guests will need to be re-registered." % len(kvm_convert))
        print print_systems(kvm_convert)

    if options.pretend:
        rhnSQL.rollback()
        sys.stderr.write("Transaction Rollbacked! \n")
    else:
        rhnSQL.commit()



def convert_system(sid, virt_type):
    get_virt_type_id_sql = "select id from rhnvirtualinstancetype where label = '%s'" % virt_type
    query = rhnSQL.prepare(get_virt_type_id_sql)
    query.execute()
    row = query.fetchone_dict() or {}
    if not row or not row.has_key('id') or not row['id']:
        print "Unknown virtual type: %s, skipping server id %s" % (virt_type,
                sid)
        return
    virt_type_id = row['id']

    get_id_sql = "SELECT sequence_nextval('rhn_vi_id_seq') as id FROM dual"
    # if oracle (sequence_nextval doesn't exist on sat 5.3)
    if not CFG.has_key('db_backend') or CFG.DB_BACKEND != 'postgresql':
        get_id_sql = "SELECT rhn_vi_id_seq.nextval as id FROM dual"
    query = rhnSQL.prepare(get_id_sql)
    query.execute()
    row = query.fetchone_dict() or {}

    if not row or not row.has_key('id') or not row['id']:
        raise 'unable to get virt instance id'

    insert_rvi_sql = "INSERT INTO rhnVirtualInstance (id, virtual_system_id, uuid) VALUES (:id, :sid, 'flex-guest')"
    insert_rvii_sql = """
                      INSERT INTO rhnVirtualInstanceInfo (instance_id, state, instance_type)
                      VALUES (:id,
                              (SELECT rvis.id FROM rhnVirtualInstanceState rvis WHERE rvis.label = 'unknown'),
                              :virt_type_id
                             )
                      """
    try:
        rhnSQL.execute(insert_rvi_sql, id=row['id'], sid=sid)
        if virt_type:
            rhnSQL.execute(insert_rvii_sql, id=row['id'],
                    virt_type_id=virt_type_id)
    except rhnSQL.SQLError, e:
        rhnSQL.rollback()
        raise rhnFault(23, str(e.args[1]), explain=0 )


def print_systems(list):
    col1 = 12
    col2 = 17
    col3 = 20
    to_ret = ""
    to_ret += "ID".ljust(col1) + "IP Address".ljust(col2) + "Name".ljust(col3) + "\n"
    to_ret +=  ("-"*2).ljust(col1) + ("-"*10).ljust(col2) + ("-"*4).ljust(col3) + "\n"
    for system in list:
        to_ret +=  str(system['server_id']).ljust(col1) + system['ipaddr'].ljust(col2) + system['name'].ljust(col3) + "\n"
    return to_ret


#returns a tuple ([],[]) of lists, first list is server_ids which can be converted to guests
#  2nd list is server_ids which are kvm guests which can't be converted
def get_guest_lists(debug=0):

    get_needed_systems_dmi_sql = """
        select SD.server_id, SD.vendor, SD.system, SD.product, SD.bios_vendor,
               SD.bios_version, SD.bios_release, SD.asset, SD.board, SN.hostname, SN.IPADDR,
               SI.checkin, S.name
        from rhnServer S inner join
             rhnServerDMI SD on S.id = SD.server_id inner join
             rhnServerNetwork SN on S.id = SN.server_id inner join
             rhnServerInfo SI on S.id = SI.server_id
        where  SD.server_id not in ( select virtual_system_id  from rhnVirtualInstance where virtual_system_id is not null )
        """

    if debug:
        Log = rhnLog.rhnLog('/var/log/rhn/update-servers.log', 5)

    dmi_query = rhnSQL.prepare(get_needed_systems_dmi_sql)

    dmi_query.execute()
    all_dmi = dmi_query.fetchall_dict()

    if not all_dmi:
        return [], []

    kvm_list = [] #kvm and guest guests, we don't know their UUID
    virt_list = [] #everything else we don't normally need their UUID
    for count, dmi in enumerate(all_dmi):
        if dmi['asset']:
            m = re.match(r".*\(system: (.*)\).*", dmi['asset'])
            if m:
                serial = m.group(1)
            else:
                serial = ''
        else:
            serial = ''
        smbios = {
            'smbios.bios.vendor': dmi['bios_vendor'] or "", #correct
            'smbios.system.serial': serial or "", #correct
            'smbios.system.manufacturer': dmi['board'] or "", # this is /dmidecode/BaseBoardInfo/Manufacturer, is it correct
            'smbios.system.product': dmi['product'] or "", #correct
        }

        (virt_type, uuid) =  parse_smbios(smbios)
        dmi['virt_type'] = virt_type
        if virt_type is None and smbios['smbios.bios.vendor'].lower() == 'xen':
            kvm_list.append(dmi)
            continue;
        if virt_type == VirtualizationType.QEMU:
            kvm_list.append(dmi)
            continue
        elif not virt_type:
            continue # not virt
        virt_list.append(dmi)
    return (virt_list, kvm_list)



if __name__ == '__main__':
    main()
