#!/usr/bin/python

import sys, os
from optparse import Option, OptionParser

sys.path.append('/usr/share/rhn')

try:
    from common import initCFG, rhnLog
    from server import rhnSQL
except ImportError:
    from spacewalk.common.rhnConfig import initCFG
    from spacewalk.common import rhnLog
    from spacewalk.server import rhnSQL

verbose = None
log = None

def message(text, l, v, stdout=False):
    if log and l:
        log.writeMessage(text)
    if (verbose and v) or stdout:
        print text


def allowed_character(c):
    if not c.isalnum() and not c in [ "_", "-"]:
        return False
    else:
        return True


def get_ks_profiles():
    sql = """
        select ksd.id as id,
               ksd.label as label,
               ksd.org_id as org_id,
               wc.name as org_name
          from rhnKSData ksd,
               web_customer wc
         where ksd.org_id = wc.id
         order by org_id, id, label
    """

    sql_h = rhnSQL.prepare(sql)
    sql_h.execute()
    return sql_h.fetchall_dict()


def list_invalid_profiles():
    profiles = get_ks_profiles()

    for profile in profiles or []:
        label_ok = True
        label_exp = ""
        for c in profile['label']:
            if not allowed_character(c):
                label_ok = False
                label_exp += "^"
            else:
                label_exp += " "

        if not label_ok:
            print "Kickstart profile: %s" % profile['label']
            print "                   %s" % label_exp
            print " * kickstart id: %s" % profile['id']
            print " * organization: %s" % profile['org_name']
            print " * org_id: %s" % profile['org_id']
            print


_update_profile_sql = """
    update rhnKSData
       set label = :label
     where id = :id
"""

def update_ks_profiles(replace):
    profiles = get_ks_profiles()
    _update_profile_sql_h = rhnSQL.prepare(_update_profile_sql)

    for profile in profiles or []:
        label_ok = True
        new_label = ""
        for c in profile['label']:
            if not allowed_character(c):
                label_ok = False
                new_label += replace
            else:
                new_label += c

        if not label_ok:
            message("Renaming profile \"%s\" -> \"%s\" (profile id: %s)" % \
                (profile['label'], new_label, profile['id']), 1, 1)

            try:
                _update_profile_sql_h.execute(label=new_label, id=profile['id'])
            except Exception, exc:
                message("Automatic rename of profile \"%s\" -> \"%s\" (profile id: %s) failed with %s" % \
                    (profile['label'], new_label, profile['id'], str(exc)), 1, 1, stdout=True)
                message("Try to rename the problematic profile in webui and continue with this script again", 1, 1, True)
                sys.exit(1)

            rhnSQL.commit()


helper_text = """
%s script is supposed to be used as a part of Red Hat Satellite
upgrade procedure and is meant to do the following:
 * list kickstart profiles with labels containing invalid characters
   (acceptable characters are: alphanumeric, '_' and '-')
 * convert kickstart profiles with label contaning invalid characters
   (it replaces invalid characters by underscore or dash symbols)

Use --help to find out more about how to use the script
"""

def main():
    global verbose, log

    basename = os.path.basename(sys.argv[0])

    if (len(sys.argv) == 1):
        print helper_text % basename
        sys.exit(0)

    options_table = [
        Option("--list-invalid-profiles", action="store_true",
            help="list kickstart profiles with invalid characters"),
        Option("--update-ks-profiles", action="store_true",
            help="update labels of kickstart profiles with invalid characters"),
        Option("--replace-dash", action="store_true",
            help="use dash symbol \"-\" to replace invalid characters"),
        Option("--replace-underscore", action="store_true",
            help="use underscore symbol \"_\" to replace invalid characters"),
        Option("--verbose", action="count", help="increase verbosity"),
        Option("--log", action="store",
			help="log the actions performed into a log file"),
    ]

    parser = OptionParser(option_list=options_table)
    (options, args) = parser.parse_args()

    if options.list_invalid_profiles and options.update_ks_profiles:
        sys.stderr.write("--list-invalid-profiles and --update-ks-profiles are mutually exclusive\n")
        sys.exit(1)

    if options.verbose:
        verbose = True

    if options.log:
        log = rhnLog.rhnLog(options.log, 5)

    initCFG('server.satellite')
    rhnSQL.initDB()

    if options.list_invalid_profiles:
        list_invalid_profiles()

    if options.update_ks_profiles:
        rc = '_'

        if options.replace_dash:
            rc = '-'
        if options.replace_underscore:
            rc = '_'

        update_ks_profiles(rc)


if __name__ == '__main__':
    main()
